scenes_list = ["ICE", "VOYAGER", "NOTHING", "EXTREME", "EARTH", "CREDITS"]
task_list = ["VisibleLight","NeutrinoVision","X-ray","xray_vision","pluto1","pluto2","gaze_voyager","gazefeet","gazeup","gazeray","gazeball","gaze_ice","vision_modules","grid","neutrino_event"]
scene_map = {"ICE":
                ["VisibleLight",
                "gazefeet",
                "gazeup"],
            "VOYAGER":
                ["VisibleLight",
                "X-ray",
                "NeutrinoVision"],
            "NOTHING":
                ["VisibleLight"],
            "EXTREME":
                ["VisibleLight",
                "X-ray",
                "NeutrinoVision"],
            "EARTH":
                ["VisibleLight"],
            "CREDITS":
                ["VisibleLight"]
            }
            